"""Main interface for the velbusaio lib."""

from __future__ import annotations

import asyncio
import logging
import pathlib
import re
import ssl
import time
from urllib.parse import urlparse

import serial
import serial_asyncio_fast

from velbusaio.channels import (
    Blind,
    Button,
    ButtonCounter,
    Channel,
    Dimmer,
    EdgeLit,
    LightSensor,
    Memo,
    Relay,
    SelectedProgram,
    Sensor,
    SensorNumber,
    Temperature,
    ThermostatChannel,
)
from velbusaio.exceptions import VelbusConnectionFailed
from velbusaio.handler import PacketHandler
from velbusaio.helpers import get_cache_dir
from velbusaio.message import Message
from velbusaio.messages.module_type_request import ModuleTypeRequestMessage
from velbusaio.messages.set_date import SetDate
from velbusaio.messages.set_daylight_saving import SetDaylightSaving
from velbusaio.messages.set_realtime_clock import SetRealtimeClock
from velbusaio.module import Module
from velbusaio.protocol import VelbusProtocol
from velbusaio.raw_message import RawMessage


class Velbus:
    """A velbus controller."""

    def __init__(
        self,
        dsn: str,
        cache_dir: str = get_cache_dir(),
        one_address: int | None = None,
    ) -> None:
        """Init the Velbus controller."""
        self._log = logging.getLogger("velbus")

        self._protocol = VelbusProtocol(
            message_received_callback=self._on_message_received,
            connection_lost_callback=self._on_connection_lost,
        )
        self._closing = False
        self._auto_reconnect = True

        self._dsn = dsn
        self._handler = PacketHandler(self, one_address)
        self._modules: dict[int, Module] = {}
        self._submodules: list[int] = []
        self._send_queue: asyncio.Queue = asyncio.Queue()
        self._cache_dir: str = cache_dir
        # make sure the cachedir exists
        pathlib.Path(self._cache_dir).mkdir(parents=True, exist_ok=True)

    def get_cache_dir(self) -> str:
        return self._cache_dir

    async def _on_message_received(self, msg: RawMessage) -> None:
        """On message received function."""
        await self._handler.handle(msg)

    def _on_connection_lost(self, exc: Exception) -> None:
        """Respond to Protocol connection lost."""
        if self._auto_reconnect and not self._closing:
            self._log.debug("Reconnecting to transport")
            asyncio.ensure_future(self.connect())

    async def add_module(
        self,
        addr: int,
        typ: int,
        serial: int | None = None,
        memorymap: int | None = None,
        build_year: int | None = None,
        build_week: int | None = None,
    ) -> None:
        """Add a found module to the module cache."""
        module = Module.factory(
            addr,
            typ,
            serial=serial,
            build_year=build_year,
            build_week=build_week,
            memorymap=memorymap,
            cache_dir=self._cache_dir,
        )
        await module.initialize(self.send)
        self._modules[addr] = module
        self._log.info(f"Found module {addr}: {module}")

    def add_submodules(self, module: Module, subList: dict[int, int]) -> None:
        """Add submodules address to module."""
        for sub_num, sub_addr in subList.items():
            if sub_addr == 0xFF:
                continue
            self._submodules.append(sub_addr)
            module._sub_address[sub_num] = sub_addr
            self._modules[sub_addr] = module
        module.cleanupSubChannels()

    def addr_is_submodule(self, addr: int) -> bool:
        """Check if an address is a submodule."""
        return addr in self._submodules

    def get_modules(self) -> dict:
        """Return the module cache."""
        return self._modules

    def get_module(self, addr: int) -> None | Module:
        """Get a module on an address."""
        if addr in self._modules:
            return self._modules[addr]
        return None

    def get_channels(self, addr: int) -> None | dict:
        """Get the channels for an address."""
        if addr in self._modules:
            return (self._modules[addr]).get_channels()
        return None

    async def stop(self) -> None:
        """Stop the controller."""
        self._closing = True
        self._auto_reconnect = False
        self._protocol.close()

    async def connect(self) -> None:
        """Connect to the bus and load all the data."""
        await self._handler.read_protocol_data()
        # connect to the bus
        if ":" in self._dsn:
            # tcp/ip combination
            if not re.search(r"^[A-Za-z0-9+.\-]+://", self._dsn):
                # if no scheme, then add the tcp://
                self._dsn = f"tcp://{self._dsn}"
            parts = urlparse(self._dsn)
            if parts.scheme == "tls":
                ctx = ssl._create_unverified_context()
            else:
                ctx = None
            try:
                (
                    _transport,
                    _protocol,
                ) = await asyncio.get_event_loop().create_connection(
                    lambda: self._protocol,
                    host=parts.hostname,
                    port=parts.port,
                    ssl=ctx,
                )

            except (ConnectionRefusedError, OSError) as err:
                raise VelbusConnectionFailed from err
        else:
            # serial port
            try:
                _transport, _protocol = (
                    await serial_asyncio_fast.create_serial_connection(
                        asyncio.get_event_loop(),
                        lambda: self._protocol,
                        url=self._dsn,
                        baudrate=38400,
                        bytesize=serial.EIGHTBITS,
                        parity=serial.PARITY_NONE,
                        stopbits=serial.STOPBITS_ONE,
                        xonxoff=0,
                        rtscts=1,
                    )
                )
            except (FileNotFoundError, serial.SerialException) as err:
                raise VelbusConnectionFailed from err

    async def start(self) -> None:
        # if auth is required send the auth key
        parts = urlparse(self._dsn)
        if parts.username:
            await self._protocol.write_auth_key(parts.username)
        # scan the bus
        await self._handler.scan()

    async def scan(self) -> None:
        """Service endpoint to restart the scan"""
        await self._handler.scan(True)

    async def sendTypeRequestMessage(self, address: int) -> None:
        msg = ModuleTypeRequestMessage(address)
        await self.send(msg)

    async def send(self, msg: Message) -> None:
        """Send a packet."""
        await self._protocol.send_message(
            RawMessage(
                priority=msg.priority,
                address=msg.address,
                rtr=msg.rtr,
                data=msg.data_to_binary(),
            )
        )

    def get_all_sensor(
        self,
    ) -> list[ButtonCounter | Temperature | LightSensor | SensorNumber]:
        return self._get_all("sensor")

    def get_all_switch(self) -> list[Relay]:
        return self._get_all("switch")

    def get_all_binary_sensor(self) -> list[Button]:
        return self._get_all("binary_sensor")

    def get_all_button(self) -> list[Button | ButtonCounter]:
        return self._get_all("button")

    def get_all_climate(self) -> list[Temperature]:
        return self._get_all("climate")

    def get_all_cover(self) -> list[Blind]:
        return self._get_all("cover")

    def get_all_select(self) -> list[SelectedProgram]:
        return self._get_all("select")

    def get_all_light(self) -> list[Dimmer]:
        return self._get_all("light")

    def get_all_led(self) -> list[Button]:
        return self._get_all("led")

    def _get_all(
        self, class_name: str
    ) -> list[
        Blind
        | Button
        | ButtonCounter
        | Sensor
        | ThermostatChannel
        | Dimmer
        | Temperature
        | SensorNumber
        | LightSensor
        | Relay
        | EdgeLit
        | Memo
        | SelectedProgram
    ]:
        """Get all channels."""
        lst = []
        for addr, mod in (self.get_modules()).items():
            if addr in self._submodules:
                continue
            for chan in (mod.get_channels()).values():
                if class_name in chan.get_categories():
                    lst.append(chan)
        return lst

    async def sync_clock(self) -> None:
        """Will send all the needed messages to sync the clock."""
        lclt = time.localtime()
        await self.send(SetRealtimeClock(wday=lclt[6], hour=lclt[3], min=lclt[4]))
        await self.send(SetDate(day=lclt[2], mon=lclt[1], year=lclt[0]))
        await self.send(SetDaylightSaving(ds=not lclt[8]))



from __future__ import annotations

from enum import IntEnum

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xD4


class CustomColorPriority(IntEnum):
    LOW_PRIORITY = 1
    MID_PRIORITY = 2
    HIGH_PRIORITY = 3


@register(COMMAND_CODE, ["VMBEL1", "VMBEL2", "VMBEL4", "VMBELO"])
class SetEdgeColorMessage(Message):
    """
    send by:
    received by: VMBEL1, VMBEL2, VMBEL4, VMBELO
    """

    apply_background_color = False
    apply_continuous_feedback_color = False
    apply_slow_blinking_feedback_color = False
    apply_fast_blinking_feedback_color = False
    custom_color_palette = False

    apply_to_left_edge = False
    apply_to_top_edge = False
    apply_to_right_edge = False
    apply_to_bottom_edge = False

    apply_to_page: int | None = None
    apply_to_all_pages = False

    background_blinking = False
    custom_color_priority: CustomColorPriority | None = None

    color_idx: int = 0

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_no_rtr(rtr)
        self.set_attributes(priority, address, rtr)

        self.apply_background_color = bool(data[0] & 0x01)
        self.apply_continuous_feedback_color = bool(data[0] & 0x02)
        self.apply_slow_blinking_feedback_color = bool(data[0] & 0x04)
        self.apply_fast_blinking_feedback_color = bool(data[0] & 0x08)
        self.custom_color_palette = bool(data[0] & 0x80)

        self.apply_to_left_edge = bool(data[1] & 0x01)
        self.apply_to_top_edge = bool(data[1] & 0x02)
        self.apply_to_right_edge = bool(data[1] & 0x04)
        self.apply_to_bottom_edge = bool(data[1] & 0x08)

        page = (data[1] & 0b0111_0000) >> 4
        if page > 0:
            self.apply_to_page = page
        self.apply_to_all_pages = bool(data[1] & 0x80)

        self.background_blinking = bool(data[2] & 0x80)

        custom_color_priority_value = data[2] & 0b0110_0000 >> 5
        if custom_color_priority_value:
            self.custom_color_priority = CustomColorPriority(
                custom_color_priority_value
            )

        self.color_idx = data[2] & 0b0001_1111

    def data_to_binary(self):
        """
        :return: bytes
        """

        byte_2 = (
            self.apply_background_color * 0x01
            + self.apply_continuous_feedback_color * 0x02
            + self.apply_slow_blinking_feedback_color * 0x04
            + self.apply_fast_blinking_feedback_color * 0x08
            + self.custom_color_palette * 0x80
        )

        byte_3 = (
            self.apply_to_left_edge * 0x01
            + self.apply_to_top_edge * 0x02
            + self.apply_to_right_edge * 0x04
            + self.apply_to_bottom_edge * 0x08
            + (((self.apply_to_page or 0) & 0xFFF) << 4)
            + self.apply_to_all_pages * 0x80
        )

        byte_4 = (
            (self.color_idx & 0xFFFF)
            + (
                (int(self.custom_color_priority) if self.custom_color_priority else 0)
                << 5
            )
            + self.background_blinking * 0x80
        )

        return bytes(
            [
                COMMAND_CODE,
                byte_2,
                byte_3,
                byte_4,
            ]
        )



import binascii
import logging
from typing import NamedTuple, Optional, Tuple

from velbusaio.const import (
    END_BYTE,
    HEADER_LENGTH,
    MAXIMUM_MESSAGE_SIZE,
    MINIMUM_MESSAGE_SIZE,
    NO_RTR,
    PRIORITIES,
    RTR,
    START_BYTE,
    TAIL_LENGTH,
)
from velbusaio.util import checksum
from velbusaio.util import checksum as calculate_checksum


class RawMessage(NamedTuple):
    priority: int
    address: int
    rtr: bool
    data: bytes

    @property
    def command(self) -> Optional[int]:
        return self.data[0] if len(self.data) > 0 else None

    @property
    def data_only(self) -> Optional[bytes]:
        return self.data[1:] if len(self.data) > 1 else None

    def to_bytes(self) -> bytes:
        # create header:
        header_bytes = bytes(
            [
                START_BYTE,
                self.priority,
                self.address,
                (RTR if self.rtr else NO_RTR) | len(self.data),
            ]
        )

        tail_bytes = bytes([checksum(header_bytes + self.data), END_BYTE])

        return header_bytes + self.data + tail_bytes

    def __repr__(self) -> str:
        return (
            f"RawMessage(priority={self.priority:02x}, address={self.address:02x},"
            f" rtr={self.rtr!r}, command={self.command},"
            f" data={binascii.hexlify(self.data, ' ')})"
        )


def create(rawmessage: bytearray) -> Tuple[Optional[RawMessage], bytearray]:
    rawmessage = _trim_buffer_garbage(rawmessage)

    while True:
        if len(rawmessage) < MINIMUM_MESSAGE_SIZE:
            logging.debug("Buffer does not yet contain a full message")
            return None, rawmessage

        try:
            return _parse(rawmessage)
        except ParseError:
            logging.error(
                f"Could not parse the message {binascii.hexlify(rawmessage)}. Truncating invalid data."
            )
            rawmessage = _trim_buffer_garbage(
                rawmessage[1:]
            )  # try to find possible start of a message


class ParseError(Exception):
    pass


def _parse(rawmessage: bytearray) -> Tuple[Optional[RawMessage], bytearray]:
    if len(rawmessage) < MINIMUM_MESSAGE_SIZE or len(rawmessage) > MAXIMUM_MESSAGE_SIZE:
        raise ValueError("Received a raw message with an illegal lemgth")
    if rawmessage[0] != START_BYTE:
        raise ValueError("Received a raw message with the wrong startbyte")

    priority = rawmessage[1]
    if priority not in PRIORITIES:
        raise ParseError(
            f"Invalid priority byte: {priority:02x} in {binascii.hexlify(rawmessage)}"
        )

    address = rawmessage[2]

    rtr = rawmessage[3] & RTR == RTR  # high nibble of the 4th byte
    data_size = rawmessage[3] & 0x0F  # low nibble of the 4th byte

    if HEADER_LENGTH + data_size + TAIL_LENGTH > len(rawmessage):
        return (
            None,
            rawmessage,
        )  # the full package is not available in the current buffer

    if rawmessage[HEADER_LENGTH + data_size + 1] != END_BYTE:
        raise ParseError(f"Invalid end byte in {binascii.hexlify(rawmessage)}")

    checksum = rawmessage[HEADER_LENGTH + data_size]

    calculated_checksum = calculate_checksum(rawmessage[: HEADER_LENGTH + data_size])

    if calculated_checksum != checksum:
        raise ParseError(
            f"Invalid checksum: expected {calculated_checksum:02x},"
            f" but got {checksum:02x} in {binascii.hexlify(rawmessage)}"
        )

    data = bytes(rawmessage[HEADER_LENGTH : HEADER_LENGTH + data_size])

    return (
        RawMessage(priority, address, rtr, data),
        rawmessage[HEADER_LENGTH + data_size + TAIL_LENGTH :],
    )


def _trim_buffer_garbage(rawmessage: bytearray) -> bytearray:
    """
    Remove leading garbage bytes from a byte stream.
    """

    # A proper message byte stream begins with 0x0F.
    if rawmessage and rawmessage[0] != START_BYTE:
        start_index = rawmessage.find(START_BYTE)
        if start_index > -1:
            #           logging.debug(
            #                "Trimming leading garbage from buffer content: {buffer} becomes {new_buffer}".format(
            #                    buffer=binascii.hexlify(rawmessage),
            #                    new_buffer=binascii.hexlify(rawmessage[start_index:]),
            #                )
            #            )
            return rawmessage[start_index:]
        else:
            logging.debug(
                "Trimming whole buffer as it does not contain the start byte: {buffer}".format(
                    buffer=binascii.hexlify(rawmessage)
                )
            )
            return []

    else:
        return rawmessage



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xA9


@register(COMMAND_CODE, ["VMBMETEO"])
class MeteoRawMessage(Message):
    """
    send by: VMBMETEO
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.rain = 0
        self.light = 0
        self.wind = 0

    def populate(self, priority, address, rtr, data):
        """
        data bytes (high + low)
            1 + 2   = current temp
            3 + 4   = min temp
            5 + 6   = max temp
        :return: None
        """
        self.needs_no_rtr(rtr)
        self.needs_data(data, 6)
        self.set_attributes(priority, address, rtr)
        self.rain = (((data[0] << 8) | data[1]) / 32) * 0.1
        self.light = ((data[2] << 8) | data[3]) / 32
        self.wind = (((data[4] << 8) | data[5]) / 32) * 0.1


@register(COMMAND_CODE, ["VMB4AN"])
class SensorRawMessage(Message):
    """
    send by: VMB4AN
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.sensor = 0
        self.mode = 0
        self.value = 0
        self.unit = None

    def populate(self, priority, address, rtr, data):
        self.needs_no_rtr(rtr)
        self.needs_data(data, 5)
        self.set_attributes(priority, address, rtr)
        self.sensor = data[0]
        self.mode = data[1]
        self.value = (data[2] << 16) | (data[3] << 8) | data[4]
        if self.mode == 0:
            self.value = self.value * 0.25
            self.unit = "mV"
        elif self.mode == 1:
            self.value = self.value * 5
            self.unit = "µA"
        elif self.mode == 2:
            self.value = self.value * 0.25
            self.unit = "ohm"
        elif self.mode == 3:
            self.value = self.value * 0.5
            self.unit = "µS"



"""
:author: Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xF1


@register(COMMAND_CODE)
class ChannelNamePart2Message(Message):
    """
    send by: VMB6IN, VMB4RYLD
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.name = ""
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        channels = self.byte_to_channels(data[0])
        self.needs_one_channel(channels)
        self.channel = channels[0]
        self.name = "".join([chr(x) for x in data[1:]])

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE, self.channels_to_byte([self.channel])]) + bytes(
            self.name, "ascii", "ignore"
        )


@register(
    COMMAND_CODE,
    [
        "VMBGP1",
        "VMBEL1",
        "VMBGP1-2",
        "VMBGP2",
        "VMBEL2",
        "VMBGP2-2",
        "VMBGP4",
        "VMBEL4",
        "VMBGP4-2",
        "VMBGPO",
        "VMBGPOD",
        "VMBGPOD-2",
        "VMBELO",
        "VMBGP4PIR",
        "VMBGP4PIR-2",
        "VMBDMI",
        "VMBDMI-R",
        "VMBIN",
        "VMBKP",
        "VMBELPIR",
        "VMBDALI",
        "VMB4AN",
        "VMB6PB-20",
        "VMBEL1-20",
        "VMBEL2-20",
        "VMBEL4-20",
        "VMBELO-20",
        "VMBGP1-20",
        "VMBGP2-20",
        "VMBGP4-20",
        "VMBGPO-20",
        "VMBDALI-20",
        "VMBEL4PIR-20",
        "VMBGP4PIR-20",
        "VMB8DC-20",
        "VMB8IN-20",
        "VMB4LEDPWM-20",
        "VMBPSUMNGR-20",
    ],
)
class ChannelNamePart2Message2(ChannelNamePart2Message):
    """
    send by: VMBGP*, VMBDALI
    received by:
    """

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.channel = data[0]
        self.name = "".join([chr(x) for x in data[1:]])


@register(COMMAND_CODE, ["VMB1BL", "VMB2BL"])
class ChannelNamePart2Message3(ChannelNamePart2Message):
    """
    send by: VMBGP*
    received by:
    """

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 5)
        self.set_attributes(priority, address, rtr)
        self.channel = (data[0] >> 1) & 0x03
        self.name = "".join([chr(x) for x in data[1:]])



#!/usr/bin/env python


class VelbusException(Exception):
    """Velbus Exception."""

    def __init__(self, value: str) -> None:
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return repr(self.value)


class VelbusConnectionFailed(VelbusException):
    def __init__(self) -> None:
        super().__init__("Connection setup failed")


class VelbusConnectionTerminated(VelbusException):
    def __init__(self) -> None:
        super().__init__("Connection terminated")



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xE6


@register(COMMAND_CODE)
class SensorTemperatureMessage(Message):
    """
    send by: VMBTS, vmbg*pd, ...
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.cur = 0
        self.min = 0
        self.max = 0

    def getCurTemp(self):
        return self.cur

    def getMaxTemp(self):
        return self.max

    def getMinTemp(self):
        return self.min

    def populate(self, priority, address, rtr, data):
        """
        data bytes (high + low)
            1 + 2   = current temp
            3 + 4   = min temp
            5 + 6   = max temp
        :return: None
        """
        self.needs_no_rtr(rtr)
        self.needs_data(data, 6)
        self.set_attributes(priority, address, rtr)
        if ((data[0] << 8) | data[1]) >> 15:
            self.cur = -127 + (((data[0] << 8) | data[1]) / 32 * 0.0625)
        else:
            self.cur = (((data[0] << 8) | data[1]) / 32) * 0.0625
        if ((data[2] << 8) | data[3]) >> 15:
            self.min = -127 + (((data[2] << 8) | data[3]) / 32 * 0.0625)
        else:
            self.min = (((data[2] << 8) | data[3]) / 32) * 0.0625
        if ((data[4] << 8) | data[5]) >> 15:
            self.max = -127 + (((data[4] << 8) | data[5]) / 32 * 0.0625)
        else:
            self.max = (((data[4] << 8) | data[5]) / 32) * 0.0625



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xBE


@register(COMMAND_CODE)
class KwhStatusMessage(Message):
    """
    send by: VMB7IN
    received by:
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channel = 0
        self.pulses = 0
        self.counter = 0
        self.kwh = 0
        self.delay = 0
        self.watt = 0

    def populate(self, priority, address, rtr, data):
        """
        -DB1    last 2 bits   = channel
        -DB1    first 6 bist  = pulses
        -DB2-5                = pulse counter
        -DB6-7                = ms/pulse
        :return: None
        """
        self.needs_no_rtr(rtr)
        self.needs_data(data, 7)
        self.set_attributes(priority, address, rtr)
        self.channel = (data[0] & 0x03) + 1
        self.pulses = (data[0] >> 2) * 100
        self.counter = (data[1] << 24) + (data[2] << 16) + (data[3] << 8) + data[4]
        self.kwh = float(float(self.counter) / self.pulses)
        self.delay = (data[5] << 8) + data[6]
        self.watt = float((1000 * 1000 * 3600) / (self.delay * self.pulses))
        if self.watt < 55:
            self.watt = 0

    def get_channels(self):
        """
        :return: list
        """
        return self.channel



"""
:author: Thomas Delaet <thomas@delaet.org> and Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xEF


@register(COMMAND_CODE)
class ChannelNameRequestMessage(Message):
    """
    send by:
    received by: VMB6IN, VMB4RYLD
    """

    def __init__(self, address=None):
        Message.__init__(self)
        self.channels = []
        self.set_defaults(address)

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 1)
        self.set_attributes(priority, address, rtr)
        self.channels = self.byte_to_channels(data[0])

    def data_to_binary(self):
        """
        :return: bytes
        """
        if isinstance(self.channels, list):
            return bytes([0xEF, self.channels_to_byte(self.channels)])
        return bytes([0xEF, 0xFF])


@register(COMMAND_CODE, ["VMB2BL"])
class ChannelNameRequestMessage2(ChannelNameRequestMessage):
    """
    send by:
    received by: VMB2BL
    """

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 1)
        self.set_attributes(priority, address, rtr)
        tmp = (data[0] >> 1) & 0x03
        self.channels = self.byte_to_channels(tmp)

    def data_to_binary(self):
        """
        :return: bytes
        """
        tmp = 0x00
        if 1 in self.channels:
            tmp += 0x03
        if 2 in self.channels:
            tmp += 0x0C
        return bytes([COMMAND_CODE, tmp])


@register(COMMAND_CODE, ["VMBDALI", "VMBDALI-20"])
class ChannelNameRequestMessage3(ChannelNameRequestMessage):
    """
    send by:
    received by: VMBDALI
    """

    def populate(self, priority, address, rtr, data):
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 1)
        self.set_attributes(priority, address, rtr)
        self.channels = data[0]

    def data_to_binary(self):
        """
        :return: bytes
        """
        return bytes([COMMAND_CODE, self.channels])



"""Constant for velbusaio.

Author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from typing import Final

PRIORITY_HIGH: Final = 0xF8
PRIORITY_FIRMWARE: Final = 0xF9
PRIORITY_LOW: Final = 0xFB
PRIORITY_THIRDPARTY: Final = 0xFA
PRIORITIES: Final = [
    PRIORITY_FIRMWARE,
    PRIORITY_HIGH,
    PRIORITY_LOW,
    PRIORITY_THIRDPARTY,
]


HEADER_LENGTH: Final = 4  # Header: [Start Byte, priority, address, RTR+data length]
TAIL_LENGTH: Final = 2  # Tail: [CRC, End Byte]
MAX_BODY_SIZE: Final = 8  # Maximum amount of data bytes in a packet

MINIMUM_MESSAGE_SIZE: Final = (
    HEADER_LENGTH + TAIL_LENGTH
)  # Smallest possible packet: [Start Byte, priority, address, RTR+data length, CRC, End Byte]
MAXIMUM_MESSAGE_SIZE: Final = MINIMUM_MESSAGE_SIZE + MAX_BODY_SIZE

START_BYTE: Final = 0x0F
END_BYTE: Final = 0x04


LENGTH_MASK: Final = 0x0F

RTR: Final = 0x40
NO_RTR: Final = 0x00

CACHEDIR: Final = ".velbuscache"

# Module scan timeout values (in mSec)
SCAN_MODULETYPE_TIMEOUT: Final = 3000  # time to wait for ModuleTypeRequest
SCAN_MODULEINFO_TIMEOUT_INITIAL: Final = 1000  # time to wait for first info (status)
SCAN_MODULEINFO_TIMEOUT_INTERVAL: Final = (
    150  # time to wait for info interval (between next message)
)

DEVICE_CLASS_ILLUMINANCE: Final = "illuminance"
DEVICE_CLASS_TEMPERATURE: Final = "temperature"
TEMP_CELSIUS: Final = "°C"
ENERGY_KILO_WATT_HOUR: Final = "kWh"
ENERGY_WATT_HOUR: Final = "Wh"
VOLUME_CUBIC_METER: Final = "m³"  # Not an official constant at HA yet
VOLUME_CUBIC_METER_HOUR: Final = "m³/h"  # Not an official constant at HA yet
VOLUME_LITERS: Final = "L"
VOLUME_LITERS_HOUR: Final = "L/h"  # Not an official constant at HA yet

CHANNEL_SELECTED_PROGRAM: Final = 96
CHANNEL_EDGE_LIT: Final = 97
CHANNEL_MEMO_TEXT: Final = 98
CHANNEL_LIGHT_VALUE: Final = 99

SLEEP_TIME = 60 / 1000



"""
:author: Maikel Punie <maikel.punie@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.message import Message

COMMAND_CODE = 0xB7


@register(COMMAND_CODE)
class SetDate(Message):
    """
    received by all modules
    """

    def __init__(self, address=0x00, day=None, mon=None, year=None) -> None:
        Message.__init__(self)
        self._day = day
        self._mon = mon
        self._year = year
        self.set_defaults(address)

    def set_defaults(self, address) -> None:
        if address is not None:
            self.set_address(address)
        self.set_low_priority()
        self.set_no_rtr()

    def populate(self, priority, address, rtr, data) -> None:
        """
        :return: None
        """
        self.needs_low_priority(priority)
        self.needs_no_rtr(rtr)
        self.needs_data(data, 4)
        self.set_attributes(priority, address, rtr)
        self._day = data[0]
        self._mon = data[1]
        self._year = (data[2] << 8) + data[3]

    def data_to_binary(self) -> bytes:
        """
        :return: bytes
        """
        return bytes(
            [
                COMMAND_CODE,
                self._day,
                self._mon,
                (self._year >> 8),
                (self._year & 0x00FF),
            ]
        )



"""
:author: David Danssaert <david.danssaert@gmail.com>
"""

from __future__ import annotations

from velbusaio.command_registry import register
from velbusaio.messages.module_status import ModuleStatusMessage

COMMAND_CODE = 0xEB


@register(COMMAND_CODE, ["VMB8IR"])
class IRReceiverStatusMessage(ModuleStatusMessage):
    """
    send by: VMB8IR
    received by:
    """



"""Command registry.

:author: Maikel Punie <maikel.punie@gmail.com> and Thomas Delaet <thomas@delaet.org>
"""

from __future__ import annotations

MODULE_DIRECTORY = {
    0x01: "VMB8PB",
    0x02: "VMB1RY",
    0x03: "VMB1BL",
    0x04: "VMBPSUMNGR-20",
    0x05: "VMB6IN",
    0x06: "VMB4LEDPWM-20",
    0x07: "VMB1DM",
    0x08: "VMB4RY",
    0x09: "VMB2BL",
    0x0A: "VMB8IR",
    0x0B: "VMB4PD",
    0x0C: "VMB1TS",
    0x0D: "VMB1TH",
    0x0E: "VMB1TC",
    0x0F: "VMB1LED",
    0x10: "VMB4RYLD",
    0x11: "VMB4RYNO",
    0x12: "VMB4DC",
    0x13: "VMBLCDWB",
    0x14: "VMBDME",
    0x15: "VMBDMI",
    0x16: "VMB8PBU",
    0x17: "VMB6PBN",
    0x18: "VMB2PBN",
    0x19: "VMB6PBB",
    0x1A: "VMB4RF",
    0x1B: "VMB1RYNO",
    0x1C: "VMB1BLE",
    0x1D: "VMB2BLE",
    0x1E: "VMBGP1",
    0x1F: "VMBGP2",
    0x20: "VMBGP4",
    0x21: "VMBGPO",
    0x22: "VMB7IN",
    0x23: "VMBPIRO-10",
    0x24: "VMBGPTC",
    0x28: "VMBGPOD",
    0x29: "VMB1RYNOS",
    0x2A: "VMBPIRM",
    0x2B: "VMBPIRC",
    0x2C: "VMBPIRO",
    0x2D: "VMBGP4PIR",
    0x2E: "VMB1BLS",
    0x2F: "VMBDMI-R",
    0x30: "VMBRFR8S",
    0x31: "VMBMETEO",
    0x32: "VMB4AN",
    0x33: "VMBVP01",
    0x34: "VMBEL1",
    0x35: "VMBEL2",
    0x36: "VMBEL4",
    0x37: "VMBELO",
    0x38: "VMBELPIR",
    0x39: "VMBSIG",
    0x3A: "VMBGP1-2",
    0x3B: "VMBGP2-2",
    0x3C: "VMBGP4-2",
    0x3D: "VMBGPOD-2",
    0x3E: "VMBGP4PIR-2",
    0x3F: "VMCM3",
    0x40: "VMBUSBIP",
    0x41: "VMB1RYS",
    0x42: "VMBKP",
    0x43: "VMBIN",
    0x44: "VMB4PB",
    0x45: "VMBDALI",
    0x48: "VMB4RYLD-10",
    0x49: "VMB4RYNO-10",
    0x4A: "VMB2BLE-10",
    0x4B: "VMB8DC-20",
    0x4C: "VMB6PB-20",
    0x4D: "VMBPIR-20",
    0x4E: "VMB8IN-20",
    0x4F: "VMBEL1-20",
    0x50: "VMBEL2-20",
    0x51: "VMBEL4-20",
    0x52: "VMBELO-20",
    0x54: "VMBGP1-20",
    0x55: "VMBGP2-20",
    0x56: "VMBGP4-20",
    0x57: "VMBGPO-20",
    0x59: "VMBPIRO-20",
    0x5A: "VMBDALI-20",
    0x5B: "VMBSIG-20",
    0x5C: "VMBEL4PIR-20",
    0x5F: "VMBGP4PIR-20",
    0x60: "VMBSIG-21",
}


class CommandRegistry:
    """Command registry class."""

    def __init__(self, module_directory: dict) -> None:
        """Init method."""
        self._module_directory = module_directory
        self._default_commands = {}
        self._overrides = {}

    def register_command(
        self, command_value: int, command_class: type, module_name: str | None = None
    ) -> None:
        """Register a command."""
        if command_value < 0 or command_value > 255:
            raise ValueError("Command_value should be >=0 and <=255")
        if module_name and module_name not in self._module_directory.values():
            raise Exception(f"Module name {module_name} not known")
        if module_name:
            module_type = next(
                (
                    mtype
                    for mtype, mname in self._module_directory.items()
                    if mname == module_name
                ),
                None,
            )
            self._register_override(command_value, command_class, module_type)
        else:
            self._register_default(command_value, command_class)

    def _register_override(
        self, command_value: int, command_class: type, module_type: str
    ) -> None:
        """Register and override."""
        if module_type not in self._overrides:
            self._overrides[module_type] = {}
        if command_value not in self._overrides[module_type]:
            self._overrides[module_type][command_value] = command_class
        else:
            raise Exception(
                f"double registration in command registry {command_value} {command_class}"
            )

    def _register_default(self, command_value: int, command_class: type) -> None:
        """Register a default command."""
        if command_value not in self._default_commands:
            self._default_commands[command_value] = command_class
        else:
            raise Exception("double registration in command registry")

    def has_command(self, command_value: int, module_type: int = 0) -> bool:
        """Find a command."""
        if module_type in self._overrides:
            if command_value in self._overrides[module_type]:
                return True
        if command_value in self._default_commands:
            return True
        return False

    def get_command(self, command_value: int, module_type: int = 0) -> None | type:
        """Search a command in the registry."""
        if module_type in self._overrides:
            if command_value in self._overrides[module_type]:
                return self._overrides[module_type][command_value]
        if command_value in self._default_commands:
            return self._default_commands[command_value]
        return None


commandRegistry = CommandRegistry(MODULE_DIRECTORY)


def register(command_value: int, module_types: list[str] | None = None):
    """Register decorator."""

    def inner_register(command_class):
        if module_types:
            for module_type in module_types:
                commandRegistry.register_command(
                    command_value, command_class, module_type
                )
        else:
            commandRegistry.register_command(command_value, command_class)
        return command_class

    return inner_register
